from airtouch2.protocol.at2.constants import CommandMessageConstants, CommandMessageType, GroupCommands, MessageLength
from airtouch2.protocol.at2.message_common import add_checksum_message_buffer
from airtouch2.common.Buffer import Buffer
from airtouch2.common.interfaces import Serializable


def prime_group_control_message_buffer(target_group: int) -> Buffer:
    buffer = Buffer(MessageLength.COMMAND)
    buffer.append_bytes(CommandMessageConstants.BYTE_0.to_bytes(1, 'little'))
    buffer.append_bytes(CommandMessageType.GROUP_CONTROL.to_bytes(1, 'little'))
    buffer.append_bytes(CommandMessageConstants.BYTE_2.to_bytes(1, 'little'))
    buffer.append_bytes(target_group.to_bytes(1, 'little'))
    return buffer


class ToggleGroup(Serializable):
    def __init__(self, target_group_number: int):
        self.target_group = target_group_number

    def to_bytes(self) -> bytes:
        buffer = prime_group_control_message_buffer(self.target_group)
        buffer.append_bytes(CommandMessageConstants.TOGGLE.to_bytes(1, 'little'))
        buffer.append_bytes(GroupCommands.TOGGLE.to_bytes(1, 'little'))
        buffer.append_bytes(bytes([0, 0, 0, 0, 0, 0]))
        add_checksum_message_buffer(buffer)
        return buffer.to_bytes()


class ChangeDamper(Serializable):
    def __init__(self, target_group_number: int, inc: bool):
        self.target_group = target_group_number
        self.inc = inc

    def to_bytes(self) -> bytes:
        buffer = prime_group_control_message_buffer(self.target_group)
        inc_dec = GroupCommands.DAMP_INC if self.inc else GroupCommands.DAMP_DEC
        buffer.append_bytes(inc_dec.to_bytes(1, 'little'))
        buffer.append_bytes(GroupCommands.CHANGE_DAMP.to_bytes(1, 'little'))
        buffer.append_bytes(bytes([0, 0, 0, 0, 0, 0]))
        add_checksum_message_buffer(buffer)
        return buffer.to_bytes()




from __future__ import annotations
from dataclasses import dataclass
from airtouch2.protocol.at2plus.conversions import setpoint_from_value, temperature_from_value, value_from_setpoint, value_from_temperature
from airtouch2.protocol.at2plus.control_status_common import CONTROL_STATUS_SUBHEADER_LENGTH, ControlStatusSubType, SubDataLength, ControlStatusSubHeader
from airtouch2.protocol.at2plus.enums import AcFanSpeed, AcMode, AcPower
from airtouch2.protocol.at2plus.message_common import AddressMsgType, Header, MessageType, add_checksum_message_buffer, prime_message_buffer
from airtouch2.common.Buffer import Buffer
from airtouch2.common.interfaces import Serializable

AC_STATUS_LENGTH = 10


@dataclass
class AcStatus(Serializable):
    id: int
    power: AcPower
    mode: AcMode
    fan_speed: AcFanSpeed
    set_point: float | None
    temperature: float | None
    turbo: bool
    bypass: bool
    spill: bool
    timer: bool
    error: int

    def to_bytes(self) -> bytes:
        buffer = Buffer(AC_STATUS_LENGTH)
        buffer.append_bytes(
            bytes([
                ((self.power << 4) | self.id),
                ((self.mode << 4) | self.fan_speed),
                value_from_setpoint(self.set_point),
                self.turbo << 3 | self.bypass << 2 | self.spill << 1 | self.timer,
            ])
        )
        buffer.append_bytes(value_from_temperature(
            self.temperature).to_bytes(2, 'big'))
        buffer.append_bytes(self.error.to_bytes(2, 'big'))
        buffer.append_bytes(bytes([0, 0]))
        return buffer.to_bytes()

    @staticmethod
    def from_bytes(repeat_data: bytes) -> AcStatus:
        """Construct an AcStatus message from its 10-byte serial data"""
        if (len(repeat_data) != AC_STATUS_LENGTH):
            raise ValueError(f"repeat_data must be {AC_STATUS_LENGTH} bytes")
        number: int = repeat_data[0] & 0x0F
        power = AcPower.from_int((repeat_data[0] & 0xF0) >> 4)
        mode = AcMode.from_int((repeat_data[1] & 0xF0) >> 4)
        fan_speed = AcFanSpeed.from_int(repeat_data[1] & 0x0F)
        set_point: float | None = setpoint_from_value(repeat_data[2])
        turbo: bool = repeat_data[3] & 8 > 0
        bypass: bool = repeat_data[3] & 4 > 0
        spill: bool = repeat_data[3] & 2 > 0
        timer: bool = repeat_data[3] & 1 > 0
        temperature: float | None = temperature_from_value(
            int.from_bytes(repeat_data[4:6], 'big'))
        error: int = int.from_bytes(repeat_data[6:8], 'big')
        return AcStatus(number, power, mode, fan_speed, set_point, temperature, turbo, bypass, spill, timer, error)

    def __repr__(self) -> str:
        return f"""
            id: {self.id}
            power: {self.power}
            mode: {self.mode}
            fan_speed: {self.fan_speed}
            set_point: {self.set_point}
            temperature: {self.temperature}
            turbo: {self.turbo}
            bypass: {self.bypass}
            spill: {self.spill}
            timer: {self.timer}
            error: {self.error}
        """


class AcStatusMessage(Serializable):
    """AcStatus Message (can be response with repeat subdata or request with empty subdata)"""
    statuses: list[AcStatus]

    def __init__(self, statuses: list[AcStatus]):
        self.statuses = statuses

    @staticmethod
    def from_bytes(subdata: bytes) -> AcStatusMessage:
        return AcStatusMessage(
            [AcStatus.from_bytes(subdata[i:i+AC_STATUS_LENGTH])
             for i in range(0, len(subdata), AC_STATUS_LENGTH)]
        )

    def to_bytes(self) -> bytes:
        subheader = ControlStatusSubHeader(ControlStatusSubType.AC_STATUS, SubDataLength(
            0, len(self.statuses), AC_STATUS_LENGTH))
        buffer = prime_message_buffer(
            Header(
                AddressMsgType.NORMAL, MessageType.CONTROL_STATUS,
                CONTROL_STATUS_SUBHEADER_LENGTH + subheader.subdata_length.total()))
        buffer.append(subheader)
        for status in self.statuses:
            buffer.append(status)
        add_checksum_message_buffer(buffer)
        return buffer.to_bytes()




import logging
from airtouch2.protocol.at2.constants import OPEN_ISSUE_TEXT
from airtouch2.protocol.at2.enums import ACBrand, ACFanSpeed
from airtouch2.protocol.at2.lookups import GATEWAYID_BRAND_LOOKUP

_LOGGER = logging.getLogger(__name__)


def fan_speed_from_val(supported_speeds: list[ACFanSpeed], speed_val: int) -> ACFanSpeed:
    if speed_val < 5:
        # Units with no Auto speed still start with low == 1
        if ACFanSpeed.AUTO not in supported_speeds:
            speed_val -= 1
        return supported_speeds[speed_val]
    else:
        return ACFanSpeed.AUTO


def val_from_fan_speed(supported_speeds: list[ACFanSpeed], speed: ACFanSpeed):
    speed_val = supported_speeds.index(speed)
    # Units with no Auto speed still start with low == 1
    if ACFanSpeed.AUTO not in supported_speeds:
        speed_val += 1
    return speed_val


def brand_from_gateway_id(gateway_id: int) -> ACBrand | None:
    if gateway_id > 0:
        if gateway_id in GATEWAYID_BRAND_LOOKUP:
            return GATEWAYID_BRAND_LOOKUP[gateway_id]
        else:
            _LOGGER.warning(
                f"AC has an unfamiliar gateway ID: {hex(gateway_id)} - " + OPEN_ISSUE_TEXT +
                "\nInclude the gateway ID shown above")
    return None



from airtouch2.protocol.at2.message_common import add_checksum_message_buffer
from airtouch2.protocol.at2.constants import CommandMessageConstants, CommandMessageType, MessageLength
from airtouch2.common.Buffer import Buffer
from airtouch2.common.interfaces import Serializable


class RequestState(Serializable):
    """ Command to request the state of the airtouch 2 system"""

    def to_bytes(self) -> bytes:
        buffer = Buffer(MessageLength.COMMAND)
        buffer.append_bytes(CommandMessageConstants.BYTE_0.to_bytes(1, 'little'))
        buffer.append_bytes(CommandMessageType.REQUEST_STATE.to_bytes(1, 'little'))
        buffer.append_bytes(CommandMessageConstants.BYTE_2.to_bytes(1, 'little'))
        buffer.append_bytes(bytes([0, 0, 0, 0, 0, 0, 0, 0, 0]))
        add_checksum_message_buffer(buffer)
        return buffer.to_bytes()




from __future__ import annotations
from enum import IntEnum


class AcSetPower(IntEnum):
    TOGGLE = 1
    OFF = 2
    ON = 3
    AWAY = 4
    SLEEP = 5
    UNCHANGED = 15

    @staticmethod
    def from_int(val: int) -> AcSetPower:
        try:
            return AcSetPower(val)
        except ValueError:
            return AcSetPower.UNCHANGED


class AcPower(IntEnum):
    OFF = 0
    ON = 1
    AWAY_OFF = 2
    AWAY_ON = 3
    SLEEP = 5
    NOT_AVAILABLE = 15

    @staticmethod
    def from_int(val: int) -> AcPower:
        try:
            return AcPower(val)
        except ValueError:
            return AcPower.NOT_AVAILABLE


class AcSetMode(IntEnum):
    AUTO = 0
    HEAT = 1
    DRY = 2
    FAN = 3
    COOL = 4
    UNCHANGED = 15

    @staticmethod
    def from_int(val: int) -> AcSetMode:
        try:
            return AcSetMode(val)
        except ValueError:
            return AcSetMode.UNCHANGED


class AcMode(IntEnum):
    AUTO = 0
    HEAT = 1
    DRY = 2
    FAN = 3
    COOL = 4
    AUTO_HEAT = 8
    AUTO_COOL = 9
    NOT_AVAILABLE = 15

    @staticmethod
    def from_int(val: int) -> AcMode:
        try:
            return AcMode(val)
        except ValueError:
            return AcMode.NOT_AVAILABLE


class AcFanSpeed(IntEnum):
    AUTO = 0
    QUIET = 1
    LOW = 2
    MEDIUM = 3
    HIGH = 4
    POWERFUL = 5
    TURBO = 6
    UNCHANGED = 15

    @staticmethod
    def from_int(val: int) -> AcFanSpeed:
        try:
            return AcFanSpeed(val)
        except ValueError:
            return AcFanSpeed.UNCHANGED

class GroupPower(IntEnum):
    OFF = 0
    ON = 1
    TURBO = 3

class GroupSetDamper(IntEnum):
    UNCHANGED = 0
    INC = 2
    DEC = 3
    SET = 4

    @staticmethod
    def from_int(val: int) -> GroupSetDamper:
        try:
            return GroupSetDamper(val)
        except ValueError:
            return GroupSetDamper.UNCHANGED

class GroupSetPower(IntEnum):
    UNCHANGED = 0
    NEXT = 1
    OFF = 2
    ON = 3
    TURBO = 5

    @staticmethod
    def from_int(val: int) -> GroupSetPower:
        try:
            return GroupSetPower(val)
        except ValueError:
            return GroupSetPower.UNCHANGED


from enum import IntEnum


class MessageLength(IntEnum):
    UNDETERMINED = 0
    COMMAND = 13
    RESPONSE = 395


class CommandMessageConstants(IntEnum):
    BYTE_0 = 85   # Byte 0 of command always fixed
    BYTE_2 = 12   # Byte 2 of command always fixed
    TOGGLE = 128  # toggle is common to both ACCommands and GroupCommands


class CommandMessageType(IntEnum):
    # Go in byte 1 of command messages
    REQUEST_STATE = 1
    GROUP_CONTROL = 129
    AC_CONTROL = 134


class ACCommands(IntEnum):
    # Go in byte 4 of CommandMessageType.AC_CONTROL messages
    SET_MODE = 129
    SET_FAN_SPEED = 130
    TEMP_DEC = 147
    TEMP_INC = 163


class GroupCommands(IntEnum):
    CHANGE_DAMP = 1  # 1 goes in byte 5 for change damp
    TOGGLE = 0       # 0 goes in byte 5 for toggle
    DAMP_DEC = 1     # Damper decrement (10% down) in byte 4
    DAMP_INC = 2     # Damper increment (10% up) in byte 4


class ResponseMessageConstants(IntEnum):
    LONG_STRING_LENGTH = 16
    SHORT_STRING_LENGTH = 8


class ResponseMessageOffsets(IntEnum):
    # Header is 2 bytes
    HEADER = 0
    # There are 16 zones
    # zone names are 8 bytes (ResponseMessageConstants.SHORT_STRING_LENGTH)
    GROUP_NAMES_START = 100
    # zone statuses are 1 byte each
    ZONE_STATUSES_START = 228
    # In order, each group and the number of zones it consists of
    # Zones must be in consecutive order
    GROUP_ZONES_START = 244
    # Zone strengths are 1 byte each
    ZONE_DAMPS_START = 276
    NUM_GROUPS = 292
    TURBO_GROUP = 297
    # Contains isTurbo, isSafety, isSpill bits of the 2 ACs
    ACs_STATUS = 299
    TOUCHPAD_TEMP = 323  # I think?? Idk what happens when there's 2 touchpads then...
    # System name is 16 bytes (ResponseMessageConstants.LONG_STRING_LENGTH)
    SYSTEM_NAME = 324
    # The following have 2 consecutuve bytes of each type, 1 for each AC.
    # On/Off, isError, 'Auto off' enabled, 'thermistor on AC', program number
    AC_STATUS_START = 354
    AC_BRAND_START = 356
    AC_MODE_START = 358
    AC_FAN_SPEED_START = 360
    AC_SET_TEMP_START = 362
    AC_MEASURED_TEMP_START = 364
    AC_ERROR_CODE_START = 366
    AC_GATEWAY_ID_START = 368
    AC_NAME_START = 370  # AC names are 8 bytes (ResponseMessageConstants.SHORT_STRING_LENGTH)
    HASH = 394


OPEN_ISSUE_TEXT = "please open an issue and detail your system:\n\thttps://github.com/nathanvdh/airtouch2-python/issues/new"



from airtouch2.protocol.at2.conversions import val_from_fan_speed
from airtouch2.protocol.at2.message_common import add_checksum_message_buffer
from airtouch2.protocol.at2.constants import ACCommands, CommandMessageConstants, CommandMessageType, MessageLength
from airtouch2.protocol.at2.enums import ACFanSpeed, ACMode
from airtouch2.common.Buffer import Buffer
from airtouch2.common.interfaces import Serializable


def prime_ac_control_message_buffer(target_ac: int) -> Buffer:
    buffer = Buffer(MessageLength.COMMAND)
    buffer.append_bytes(CommandMessageConstants.BYTE_0.to_bytes(1, 'little'))
    buffer.append_bytes(CommandMessageType.AC_CONTROL.to_bytes(1, 'little'))
    buffer.append_bytes(CommandMessageConstants.BYTE_2.to_bytes(1, 'little'))
    buffer.append_bytes(target_ac.to_bytes(1, 'little'))
    return buffer


class ChangeSetTemperature(Serializable):
    def __init__(self, target_ac_number: int, inc: bool):
        self.target_ac = target_ac_number
        self.inc = inc

    def to_bytes(self) -> bytes:
        buffer = prime_ac_control_message_buffer(self.target_ac)
        inc_dec = ACCommands.TEMP_INC if self.inc else ACCommands.TEMP_DEC
        buffer.append_bytes(inc_dec.to_bytes(1, 'little'))
        buffer.append_bytes(bytes([0, 0, 0, 0, 0, 0, 0]))
        add_checksum_message_buffer(buffer)
        return buffer.to_bytes()


class ToggleAc(Serializable):
    def __init__(self, target_ac_number: int):
        self.target_ac = target_ac_number

    def to_bytes(self) -> bytes:
        buffer = prime_ac_control_message_buffer(self.target_ac)
        buffer.append_bytes(CommandMessageConstants.TOGGLE.to_bytes(1, 'little'))
        buffer.append_bytes(bytes([0, 0, 0, 0, 0, 0, 0]))
        add_checksum_message_buffer(buffer)
        return buffer.to_bytes()


class SetFanSpeed(Serializable):
    def __init__(self, target_ac_number: int, supported_fan_speeds: list[ACFanSpeed], fan_speed: ACFanSpeed):
        self.target_ac = target_ac_number
        self.fan_speed_val: int = val_from_fan_speed(supported_fan_speeds, fan_speed)

    def to_bytes(self) -> bytes:
        buffer = prime_ac_control_message_buffer(self.target_ac)
        buffer.append_bytes(ACCommands.SET_FAN_SPEED.to_bytes(1, 'little'))
        buffer.append_bytes(self.fan_speed_val.to_bytes(1, 'little'))
        buffer.append_bytes(bytes([0, 0, 0, 0, 0, 0]))
        add_checksum_message_buffer(buffer)
        return buffer.to_bytes()


class SetMode(Serializable):
    def __init__(self, target_ac_number: int, mode: ACMode):
        self.target_ac = target_ac_number
        self.mode = mode

    def to_bytes(self) -> bytes:
        buffer = prime_ac_control_message_buffer(self.target_ac)
        buffer.append_bytes(ACCommands.SET_MODE.to_bytes(1, 'little'))
        buffer.append_bytes(self.mode.to_bytes(1, 'little'))
        buffer.append_bytes(bytes([0, 0, 0, 0, 0, 0]))
        add_checksum_message_buffer(buffer)
        return buffer.to_bytes()



from __future__ import annotations
from dataclasses import dataclass

from airtouch2.common.interfaces import Serializable
from airtouch2.protocol.at2plus.control_status_common import CONTROL_STATUS_SUBHEADER_LENGTH, ControlStatusSubHeader, ControlStatusSubType, SubDataLength
from airtouch2.protocol.at2plus.enums import GroupPower
from airtouch2.common.Buffer import Buffer
from airtouch2.protocol.at2plus.message_common import AddressMsgType, Header, MessageType, add_checksum_message_buffer, prime_message_buffer

GROUP_STATUS_LENGTH = 8


@dataclass
class GroupStatus(Serializable):
    id: int
    power: GroupPower
    damp: int
    supports_turbo: bool
    spill_active: bool

    def to_bytes(self) -> bytes:
        buffer = Buffer(GROUP_STATUS_LENGTH)
        buffer.append_bytes(
            bytes([
                (self.power << 6) | self.id,
                self.damp,
                0, 0, 0, 0,
                self.supports_turbo << 7 | self.spill_active << 1,
                0
            ])
        )
        return buffer.to_bytes()

    @staticmethod
    def from_bytes(repeat_data: bytes) -> GroupStatus:
        if (len(repeat_data) != GROUP_STATUS_LENGTH):
            raise ValueError(f"repeat_data must be {GROUP_STATUS_LENGTH} bytes")
        id = repeat_data[0] & 0x3F
        power = GroupPower((repeat_data[0] >> 6) & 3)
        damp = repeat_data[1] & 0x7F
        supports_turbo = ((repeat_data[6] >> 7) & 1) > 0
        spill_active = ((repeat_data[6] >> 1) & 1) > 0

        return GroupStatus(id, power, damp, supports_turbo, spill_active)

    def __repr__(self) -> str:
        return f"""
            id: {self.id}
         power: {self.power}
        damper: {self.damp}%
supports_turbo: {self.supports_turbo}
  spill_active: {self.spill_active}"""


class GroupStatusMessage(Serializable):
    """GroupStatus message (can be response with repeat subdata or request with empty subdata)"""
    statuses: list[GroupStatus]

    def __init__(self, statuses: list[GroupStatus]):
        self.statuses = statuses

    @staticmethod
    def from_bytes(subdata: bytes) -> GroupStatusMessage:
        return GroupStatusMessage(
            [GroupStatus.from_bytes(subdata[i:i+GROUP_STATUS_LENGTH])
             for i in range(0, len(subdata), GROUP_STATUS_LENGTH)]
        )

    def to_bytes(self) -> bytes:
        subheader = ControlStatusSubHeader(ControlStatusSubType.GROUP_STATUS,
                                           SubDataLength(0, len(self.statuses),
                                                         GROUP_STATUS_LENGTH))
        buffer = prime_message_buffer(
            Header(
                AddressMsgType.NORMAL, MessageType.CONTROL_STATUS,
                CONTROL_STATUS_SUBHEADER_LENGTH + subheader.subdata_length.total()))
        buffer.append(subheader)
        for status in self.statuses:
            buffer.append(status)
        add_checksum_message_buffer(buffer)
        return buffer.to_bytes()



from __future__ import annotations

import logging
from dataclasses import dataclass
from itertools import compress
from pprint import pformat

from airtouch2.protocol.at2.constants import OPEN_ISSUE_TEXT, MessageLength, ResponseMessageConstants, ResponseMessageOffsets
from airtouch2.protocol.at2.conversions import brand_from_gateway_id, fan_speed_from_val
from airtouch2.protocol.at2.enums import ACBrand, ACFanSpeed, ACMode

_LOGGER = logging.getLogger(__name__)


def _resolve_brand(gateway_id: int, reported_brand: int) -> ACBrand:
    # Brand based on gateway ID takes priority according to app smali
    gateway_based_brand = brand_from_gateway_id(gateway_id)
    return gateway_based_brand if gateway_based_brand is not None else ACBrand(reported_brand)


def _parse_name(name: bytes) -> str:
    # There's probably a better way of doing this.
    return name.decode().split()[0].split("\0")[0]

# TODO: read through app smali code more and do this more properly
# Currently I'm assuming:
#   4 speed is always LOW, MED, HIGH, POWERFUL (EXCEPT for fujitsus with 4 speeds)
#   3 speed is always LOW, MED, HIGH
#   2 speed is always LOW, HIGH
#   Daikins have no Auto
#   Gateway ID 0x14 has no Auto
#   Gateway IDs 0xFF with 3 speed have no Auto
# This is based on the app decompiled code


def _supported_fan_speeds(brand: ACBrand, num_supported_speeds: int, gateway_id: int) -> list[ACFanSpeed]:
    all_speeds: list[ACFanSpeed] = list(ACFanSpeed.__members__.values())
    supported_speeds: list[ACFanSpeed] = []

    if brand == ACBrand.FUJITSU and num_supported_speeds == 4:
        supported_speeds = all_speeds[:5]
        return supported_speeds

    # Check cases that don't support Auto mode
    if (brand == ACBrand.DAIKIN or (gateway_id == 0xFF and num_supported_speeds == 3) or gateway_id == 0x14):
        supported_speeds = []
    else:
        supported_speeds = [ACFanSpeed.AUTO]

    if num_supported_speeds > 2:
        supported_speeds += all_speeds[2:2+num_supported_speeds]
    elif num_supported_speeds == 2:
        supported_speeds += [ACFanSpeed.LOW, ACFanSpeed.HIGH]
    elif num_supported_speeds < 2:
        _LOGGER.warning(
            f"AC reports less than 2 supported fan speeds, this is unusual - " + OPEN_ISSUE_TEXT)

    return supported_speeds


@dataclass
class AcInfo:
    number: int  # 0 or 1
    name: str
    active: bool
    mode: ACMode
    supported_fan_speeds: list[ACFanSpeed]
    fan_speed: ACFanSpeed
    set_temp: int
    measured_temp: int
    brand: ACBrand
    program: int
    error: bool
    error_code: int
    thermistor: bool
    turbo: bool
    safety: bool
    spill: bool

    def _get_status_strings(self):
        flags = [self.error, self.safety, self.spill, self.turbo]
        flag_names = ['ERROR', 'SAFETY', 'SPILL', 'TURBO']
        statuses = list(compress(flag_names, flags))
        if not statuses:
            statuses.append('NORMAL')
        return statuses

    def __str__(self):
        return f"""
        AC Name:\t\t{self.name}
        AC Number:\t\t{self.number}
        Active:\t\t\t{self.active}
        Status:\t\t\t{self._get_status_strings()}
        Mode:\t\t\t{self.mode}
        Fan Speed:\t\t{self.fan_speed}
        Supported Speeds:\t{[s.name for s in self.supported_fan_speeds]}
        Measured Temp:\t\t{self.measured_temp}
        Set Temp:\t\t{self.set_temp}
        Brand:\t\t\t{self.brand}
        Error Code:\t\t{self.error_code}
        """

    @staticmethod
    def parse(ac_number: int, ac_status1: int, ac_error_code: int, ac_status2: int, ac_mode: int, ac_fan_speed: int,
              ac_set_temp: int, ac_measured_temp: int, ac_brand: int, ac_gateway_id: int, ac_name: bytes) -> AcInfo | None:

        if (ac_fan_speed == 0 and ac_fan_speed == 0 and ac_gateway_id == 0):
            # This AC isn't connected
            return None

        # MS bit is on/off
        active = (ac_status1 & 0x80 > 0)
        # Bit 7 is whether AC is errored
        error = (ac_status1 & 0x40 > 0)
        # bit 4 is whether or not 'thermistor on AC' is checked (0 = checked)
        thermistor = (ac_status1 & 0x04 == 0)
        # lowest 3 bits are AC program number
        program = (ac_status1 & 0x07)

        # 6, 5 is turbo for AC0 and AC1
        turbo = (ac_status2 & (1 << (5 - ac_number))) > 0
        # 4, 3 is safety
        safety = (ac_status2 & (1 << (3 - ac_number))) > 0
        # 2, 1 is safety
        spill = (ac_status2 & (1 << (1 - ac_number))) > 0

        # simply 0-4, see ACMode enum
        mode = ACMode(ac_mode)

        brand = _resolve_brand(ac_gateway_id, ac_brand)

        # most signficant byte is # of speeds, least significant byte is speed (the meaning of which depends), see AT2Aircon::_set_true_and_supported_fan_speed()
        num_fan_speeds = (ac_fan_speed & 0xF0) >> 4
        supported_fan_speeds = _supported_fan_speeds(brand, num_fan_speeds, ac_gateway_id)

        fan_speed = fan_speed_from_val(supported_fan_speeds, ac_fan_speed & 0x0F)

        name = _parse_name(ac_name)

        return AcInfo(
            ac_number, name, active, mode, supported_fan_speeds, fan_speed, ac_set_temp, ac_measured_temp, brand,
            program, error, ac_error_code, thermistor, turbo, safety, spill)


@dataclass
class ZoneInfo:
    active: bool
    spill: bool
    damp: int

    @staticmethod
    def parse(damp: int, zone_status: int) -> ZoneInfo:
        return ZoneInfo(zone_status & 0x80 > 0, zone_status & 0x40 > 0, damp)


@dataclass
class GroupInfo:
    name: str
    number: int
    active: bool
    damp: int
    spill: bool
    turbo: bool

    def _get_status_strings(self):
        flags = [self.spill, self.turbo]
        flag_names = ['SPILL', 'TURBO']
        statuses = list(compress(flag_names, flags))
        if not statuses:
            statuses.append('NORMAL')
        return statuses

    def __str__(self):
        return f"""
        Group Name:\t{self.name}
        Group Number:\t{self.number}
        Active:\t\t{self.active}
        Status:\t\t{self._get_status_strings()}
        Damper:\t\t{f'{self.damp*10}%'}
        """


@dataclass
class SystemInfo:
    """ The state of the airtouch2 system"""
    aircons_by_id: dict[int, AcInfo]
    groups_by_id: dict[int, GroupInfo]
    touchpad_temp: int
    system_name: str

    # TODO: 'Bufferise' this code?
    @staticmethod
    def from_bytes(raw_response: bytes) -> SystemInfo:
        assert len(raw_response) == MessageLength.RESPONSE, f"Response message must be {MessageLength.RESPONSE} bytes"

        # ACs

        aircons_by_id: dict[int, AcInfo] = {}
        for ac_id in range(2):
            name_start = ResponseMessageOffsets.AC_NAME_START + ac_id * ResponseMessageConstants.SHORT_STRING_LENGTH
            name_end = name_start + ResponseMessageConstants.SHORT_STRING_LENGTH

            ac_info = AcInfo.parse(ac_id,
                                   raw_response[ResponseMessageOffsets.AC_STATUS_START + ac_id],
                                   raw_response[ResponseMessageOffsets.AC_ERROR_CODE_START + ac_id],
                                   raw_response[ResponseMessageOffsets.ACs_STATUS + ac_id],
                                   raw_response[ResponseMessageOffsets.AC_MODE_START + ac_id],
                                   raw_response[ResponseMessageOffsets.AC_FAN_SPEED_START + ac_id],
                                   raw_response[ResponseMessageOffsets.AC_SET_TEMP_START + ac_id],
                                   raw_response[ResponseMessageOffsets.AC_MEASURED_TEMP_START + ac_id],
                                   raw_response[ResponseMessageOffsets.AC_BRAND_START + ac_id],
                                   raw_response[ResponseMessageOffsets.AC_GATEWAY_ID_START + ac_id],
                                   raw_response[name_start:name_end])
            if ac_info is not None:
                aircons_by_id[ac_id] = ac_info

        # Groups
        # TODO: Factor out into an analogous 'GroupInfo.parse()'

        num_groups = raw_response[ResponseMessageOffsets.NUM_GROUPS]
        turbo_group = raw_response[ResponseMessageOffsets.TURBO_GROUP]
        groups_by_id: dict[int, GroupInfo] = {}

        for group_id in range(num_groups):
            name_start = ResponseMessageOffsets.GROUP_NAMES_START + group_id * ResponseMessageConstants.SHORT_STRING_LENGTH
            name_end = name_start + ResponseMessageConstants.SHORT_STRING_LENGTH
            name = _parse_name(raw_response[name_start:name_end])

            group_zones = raw_response[ResponseMessageOffsets.GROUP_ZONES_START + group_id]
            start_zone = (group_zones & 0xF0) >> 4
            num_zones = group_zones & 0x0F
            
            zone_info = ZoneInfo.parse(raw_response[ResponseMessageOffsets.ZONE_DAMPS_START + start_zone],
                                       raw_response[ResponseMessageOffsets.ZONE_STATUSES_START + start_zone])
            active = zone_info.active
            spill = zone_info.spill
            damp = zone_info.damp

            mismatches: set[str] = set()
            for zone_number in range(start_zone+1, start_zone + num_zones):
                zone_info = ZoneInfo.parse(raw_response[ResponseMessageOffsets.ZONE_DAMPS_START + zone_number],
                                           raw_response[ResponseMessageOffsets.ZONE_STATUSES_START + zone_number])
                # this group is spilling if any of its zones are
                if not spill:
                    spill = zone_info.spill
                # these should match for all zones that comprise this group
                if (damp != zone_info.damp):
                    mismatches.add("damper percents")
                if (active != zone_info.active):
                    mismatches.add("on/off states")
                if mismatches:
                    _LOGGER.warning(f"Zones of group '{name}' have mismatching {', '.join(mismatches)}")

            turbo = True if turbo_group == group_id else False

            groups_by_id[group_id] = GroupInfo(name, group_id, active, damp, spill, turbo)

        # System-wide

        touchpad_temp = raw_response[ResponseMessageOffsets.TOUCHPAD_TEMP]
        system_name = raw_response[ResponseMessageOffsets.SYSTEM_NAME: ResponseMessageOffsets.SYSTEM_NAME +
                                   ResponseMessageConstants.LONG_STRING_LENGTH].decode().split("\0")[0]

        return SystemInfo(aircons_by_id, groups_by_id, touchpad_temp, system_name)

    def __str__(self):
        return f"""
        System Name:\t{self.system_name}
        ACs:\n{pformat(self.aircons_by_id)}
        Groups:\n{pformat(self.groups_by_id)}
        Touchpad Temp:\t{self.touchpad_temp}
        """



from __future__ import annotations
from typing import TYPE_CHECKING

from airtouch2.common.interfaces import Callback
from airtouch2.protocol.at2plus.enums import GroupPower, GroupSetDamper, GroupSetPower
from airtouch2.protocol.at2plus.messages.GroupControl import GroupControlMessage, GroupSettings
from airtouch2.protocol.at2plus.messages.GroupStatus import GroupStatus

if TYPE_CHECKING:
    from airtouch2.at2plus.At2PlusClient import At2PlusClient


class At2PlusGroup:
    """
    A class that represents a single airtouch2+ group.

    """

    def __init__(self, status: GroupStatus, client: At2PlusClient):
        self.status = status
        self.name: str | None = None
        self._client = client
        self._callbacks: list[Callback] = []

    async def _set_power(self, power: GroupSetPower, damp: int | None = None):
        settings = GroupSettings(self.status.id, GroupSetDamper.UNCHANGED, power, damp)
        await self._client.send(GroupControlMessage([settings]))

    async def turn_on(self, damp: int | None = None):
        await self._set_power(GroupSetPower.ON, damp)

    async def turn_off(self):
        await self._set_power(GroupSetPower.OFF)

    def is_on(self) -> bool:
        return self.status.power != GroupPower.OFF

    async def set_damp(self, new_damp: int):
        settings = GroupSettings(self.status.id, GroupSetDamper.SET, GroupSetPower.UNCHANGED, new_damp)
        await self._client.send(GroupControlMessage([settings]))

    async def set_turbo(self):
        settings = GroupSettings(self.status.id, GroupSetDamper.UNCHANGED, GroupSetPower.TURBO)
        await self._client.send(GroupControlMessage([settings]))

    def add_callback(self, callback: Callback) -> Callback:
        self._callbacks.append(callback)

        def remove_callback() -> None:
            if callback in self._callbacks:
                self._callbacks.remove(callback)

        return remove_callback

    def _update_status(self, status: GroupStatus):
        self.status = status
        for callback in self._callbacks:
            callback()

    def _update_name(self, name: str):
        self.name = name
        for callback in self._callbacks:
            callback()

    def __repr__(self):
        return str(self.status) + f"""
          name: {self.name if self.name is not None else f"Group {self.status.id}"}
"""


from airtouch2.common.interfaces import Serializable
from airtouch2.protocol.at2plus.extended_common import EXTENDED_SUBHEADER_LENGTH, ExtendedMessageSubType, ExtendedSubHeader
from airtouch2.protocol.at2plus.message_common import AddressMsgType, Header, MessageType, add_checksum_message_buffer, prime_message_buffer


def group_names_from_subdata(subdata: bytes) -> dict[int, str]:
    return {subdata[i]: subdata[i+1:i+9].decode('ascii').split("\x00")[0] for i in range(0, len(subdata), 9)}


class RequestGroupNamesMessage(Serializable):

    def to_bytes(self) -> bytes:
        buffer = prime_message_buffer(
            Header(AddressMsgType.EXTENDED, MessageType.EXTENDED, EXTENDED_SUBHEADER_LENGTH))
        buffer.append(ExtendedSubHeader(ExtendedMessageSubType.GROUP_NAME))
        add_checksum_message_buffer(buffer)
        return buffer.to_bytes()



from airtouch2.at2plus.At2PlusClient import At2PlusClient
from airtouch2.at2plus.At2PlusAircon import At2PlusAircon
from airtouch2.at2plus.At2PlusGroup import At2PlusGroup
from airtouch2.protocol.at2plus.enums import AcSetMode, AcMode, AcFanSpeed


from __future__ import annotations
from dataclasses import dataclass
from enum import IntEnum
import logging

from airtouch2.common.Buffer import Buffer
from airtouch2.protocol.at2plus.crc16_modbus import crc16
from airtouch2.common.interfaces import Serializable

# Message ID can be whatever
MESSAGE_ID = 1
HEADER_MAGIC = 0x55
HEADER_LENGTH = 8
NON_DATA_LENGTH = 10

_LOGGER = logging.getLogger(__name__)


class CommonMessageOffsets(IntEnum):
    # Header is 0x55 0x55
    HEADER = 0
    # 0x80 0xb0 when sending, 0x80 0x80 for receiving
    # Except extended message which is 0x90 0xb0 when sending and 0x90 0x90 when receiving
    ADDRESS = 2
    # 'Can be any data' - response will match
    MESAGE_ID = 4
    # Usually 0xC0 but 0x1F for extended message
    MESSAGE_TYPE = 5
    # Number of bytes of data
    DATA_LENGTH = 6
    DATA = 8
    # Data length depends on message
    # Followed by 2-byte CRC16 MODBUS of message *without* header (AKA CRC-16-ANSI AKA CRC-16-IBM)


class AddressSource(IntEnum):
    SELF = 0xB0
    OTHER = 0x9F


class AddressMsgType(IntEnum):
    UNSET = 0
    NORMAL = 0x80
    EXTENDED = 0x90


class MessageType(IntEnum):
    UNSET = 0
    CONTROL_STATUS = 0xC0
    EXTENDED = 0x1F


class Header(Serializable):
    address_msg_type: AddressMsgType
    type: MessageType
    data_length: int
    _received: bool

    def __init__(self, address_msg_type: AddressMsgType, type: MessageType, data_length: int, _received=False):
        self.address_msg_type = address_msg_type
        self.type = type
        self.data_length = data_length
        self._received = _received

    @staticmethod
    def from_bytes(header_bytes: bytes) -> Header:
        if len(header_bytes) != HEADER_LENGTH:
            raise ValueError("Unexpected header size")
        for b in header_bytes[CommonMessageOffsets.HEADER:CommonMessageOffsets.ADDRESS]:
            if (b != HEADER_MAGIC):
                raise ValueError("Message header magic is invalid")
        try:
            type = MessageType(header_bytes[CommonMessageOffsets.MESSAGE_TYPE])
        except ValueError as e:
            _LOGGER.warning(
                f"Unknown message type in header ({hex(header_bytes[CommonMessageOffsets.MESSAGE_TYPE])})", exc_info=e)
            type = MessageType.UNSET
        address_src = AddressSource(header_bytes[CommonMessageOffsets.ADDRESS])
        address_msg_type = AddressMsgType(header_bytes[CommonMessageOffsets.ADDRESS+1])
        if type == MessageType.CONTROL_STATUS:
            if (address_msg_type != AddressMsgType.NORMAL):
                raise ValueError(f"Message address value is invalid: {header_bytes.hex(':')}")
        elif type == MessageType.EXTENDED:
            if (address_msg_type != AddressMsgType.EXTENDED):
                raise ValueError(f"Message address value is invalid: {header_bytes.hex(':')}")
        id = header_bytes[CommonMessageOffsets.MESAGE_ID]
        data_length = int.from_bytes(
            header_bytes[CommonMessageOffsets.DATA_LENGTH:CommonMessageOffsets.DATA], 'big')
        return Header(address_msg_type, type, data_length, True)

    def to_bytes(self) -> bytes:
        return bytes(
            [HEADER_MAGIC, HEADER_MAGIC]) + (
            bytes([AddressSource.SELF, self.address_msg_type])
            if self._received else bytes([self.address_msg_type, AddressSource.SELF])) + bytes(
            [MESSAGE_ID, self.type]) + self.data_length.to_bytes(
            2, 'big')


def prime_message_buffer(header: Header) -> Buffer:
    buffer = Buffer(header.data_length + NON_DATA_LENGTH)
    buffer.append(header)
    return buffer


def add_checksum_message_buffer(buffer: Buffer) -> None:
    buffer.append_bytes(crc16(buffer._data[2:-2]))


def add_checksum_message_bytes(data: bytearray) -> None:
    checksum = crc16(data[2:-2])
    data[-2] = checksum[0]
    data[-1] = checksum[1]


@dataclass
class Message:
    header: Header
    data_buffer: Buffer



from __future__ import annotations
from typing import TYPE_CHECKING
from airtouch2.protocol.at2.messages.SystemInfo import GroupInfo
from airtouch2.protocol.at2.messages import ChangeDamper, ToggleGroup
from airtouch2.common.interfaces import Publisher, Callback, add_callback
if TYPE_CHECKING:
    from airtouch2.at2.At2Client import At2Client


class At2Group(Publisher):
    info: GroupInfo

    def __init__(self, client: At2Client, info: GroupInfo):
        self.info = info

        self._client = client
        self._callbacks: list[Callback] = []

    def update(self, status: GroupInfo):
        self.info = status

        for func in self._callbacks:
            func()

    def add_callback(self, callback: Callback) -> Callback:
        return add_callback(callback, self._callbacks)

    async def inc_dec_damp(self, inc: bool):
        await self._client.send(ChangeDamper(self.info.number, inc))

    async def set_damp(self, new_damp: int):
        if new_damp < 0 or new_damp > 10:
            raise ValueError("Dampers can only be set from 0 to 10")
        # Set to 0 is equivalent to turning off
        if new_damp == 0:
            await self.turn_off()
        else:
            await self.turn_on()
            damp_diff = new_damp - self.info.damp
            inc = damp_diff > 0
            for i in range(abs(damp_diff)):
                await self.inc_dec_damp(inc)

    async def _turn_on_off(self, on: bool):
        if self.info.active != on:
            await self._client.send(ToggleGroup(self.info.number))

    async def turn_off(self):
        await self._turn_on_off(False)

    async def turn_on(self):
        await self._turn_on_off(True)

    def __str__(self):
        return str(self.info)




from __future__ import annotations
from dataclasses import dataclass
from enum import IntEnum
from airtouch2.protocol.at2plus.enums import AcFanSpeed, AcSetMode
from airtouch2.protocol.at2plus.extended_common import EXTENDED_SUBHEADER_LENGTH, ExtendedMessageSubType, ExtendedSubHeader
from airtouch2.protocol.at2plus.message_common import AddressMsgType, Header, MessageType, add_checksum_message_buffer, prime_message_buffer
from airtouch2.common.interfaces import Serializable


class AcAbilitySubDataLength(IntEnum):
    V1 = 24
    V1_1 = 26


@dataclass
class SetpointLimits:
    min: int
    max: int

    def __repr__(self) -> str:
        return f"""
        min: {self.min}
        max: {self.min}
        """


@dataclass
class DualSetpointLimits:
    cool: SetpointLimits
    heat: SetpointLimits

    def __repr__(self) -> str:
        return f"""
        cool: {self.cool}
        heat: {self.heat}
        """


@dataclass
class AcAbility(Serializable):
    ac_id: int
    name: str
    start_group: int
    group_count: int
    supported_modes: list[AcSetMode]
    supported_fan_speeds: list[AcFanSpeed]
    setpoint_limits: SetpointLimits | DualSetpointLimits

    @staticmethod
    def from_bytes(data: bytes) -> AcAbility:
        if len(data) != AcAbilitySubDataLength.V1 and len(data) != AcAbilitySubDataLength.V1_1:
            raise ValueError(
                f"Invalid AcAbility length, should be {AcAbilitySubDataLength.V1} or {AcAbilitySubDataLength.V1_1}, got: {len(data)}")
        ac_id = data[0]
        following_data_length = data[1]
        if following_data_length != len(data) - 2:
            raise ValueError(
                f"Data length specified in message does not match received data length: specified {following_data_length}, got {len(data) - 2}")

        name = data[2:18].decode('ascii').split("\x00")[0]
        start_group = data[18]
        group_count = data[19]
        supported_modes = []
        for i in range(5):
            # this loop exploits the fact that the support bits are in the same order as the enum values
            if (data[20] & (1 << i)) > 0:
                supported_modes.append(AcSetMode.from_int(i))
        supported_fan_speeds = []
        for i in range(7):
            # ditto
            if (data[21] & (1 << i)) > 0:
                supported_fan_speeds.append(AcFanSpeed.from_int(i))

        set_point_limits = SetpointLimits(data[22], data[23])
        if len(data) == AcAbilitySubDataLength.V1_1:
            set_point_limits = DualSetpointLimits(set_point_limits, SetpointLimits(data[24], data[25]))

        return AcAbility(
            ac_id, name, start_group, group_count, supported_modes, supported_fan_speeds, set_point_limits)

    def to_bytes(self) -> bytes:
        data = bytes([self.ac_id, (AcAbilitySubDataLength.V1_1 - 2) if isinstance(self.setpoint_limits, DualSetpointLimits) else (
            AcAbilitySubDataLength.V1 - 2)]) + self.name.encode("ascii") + bytes(16-len(self.name)) + bytes([self.start_group, self.group_count])
        supported_modes_val: int = 0
        for mode in self.supported_modes:
            supported_modes_val |= 1 << mode
        supported_speeds_val: int = 0
        for speed in self.supported_fan_speeds:
            supported_speeds_val |= 1 << speed
        data += bytes([supported_modes_val, supported_speeds_val])
        if isinstance(self.setpoint_limits, DualSetpointLimits):
            data += bytes([self.setpoint_limits.cool.min, self.setpoint_limits.cool.max,
                          self.setpoint_limits.heat.min, self.setpoint_limits.heat.max])
        elif isinstance(self.setpoint_limits, SetpointLimits):
            data += bytes([self.setpoint_limits.min, self.setpoint_limits.max])

        return data

    def __repr__(self) -> str:
        return f"""
        id: {self.ac_id}
        name: {self.name}
        start_group: {self.start_group}
        group_count: {self.group_count}
        supported_modes: {self.supported_modes}
        supported_fan_speeds: {self.supported_fan_speeds}
        setpoint_limits: {self.setpoint_limits}
        """


class AcAbilityMessage(Serializable):
    abilities: list[AcAbility]

    def __init__(self, abilities: list[AcAbility]):
        self.abilities = abilities

    @staticmethod
    def from_bytes(subdata: bytes) -> AcAbilityMessage:
        ac_ability_list = []
        # peek the protocol version, assume they all match
        length = AcAbilitySubDataLength(subdata[1] + 2)
        # Iterate over subdata in steps of AC_ABILITY_LENGTH
        for i in range(0, len(subdata), length):
            subdata_slice = subdata[i:i+length]
            ac_ability = AcAbility.from_bytes(subdata_slice)
            ac_ability_list.append(ac_ability)
            break  # ! Ignores the other ACs in the list for now
        ac_ability_message = AcAbilityMessage(ac_ability_list)
        return ac_ability_message

    def to_bytes(self) -> bytes:
        length = AcAbilitySubDataLength.V1 if isinstance(
            self.abilities[0].setpoint_limits, SetpointLimits) else AcAbilitySubDataLength.V1_1
        buffer = prime_message_buffer(
            Header(
                AddressMsgType.EXTENDED, MessageType.EXTENDED, EXTENDED_SUBHEADER_LENGTH +
                length * len(self.abilities)))
        buffer.append(ExtendedSubHeader(ExtendedMessageSubType.ABILITY))
        for ability in self.abilities:
            buffer.append(ability)
        add_checksum_message_buffer(buffer)
        return buffer.to_bytes()


class RequestAcAbilityMessage(Serializable):
    ac_id: int | None

    def __init__(self, ac_id: int | None = None):
        self.ac_id = ac_id

    def to_bytes(self) -> bytes:
        buffer = prime_message_buffer(
            Header(
                AddressMsgType.EXTENDED, MessageType.EXTENDED, EXTENDED_SUBHEADER_LENGTH +
                (1 if self.ac_id is not None else 0)))
        buffer.append(ExtendedSubHeader(ExtendedMessageSubType.ABILITY))
        if self.ac_id is not None:
            buffer.append_bytes(bytes([self.ac_id]))
        add_checksum_message_buffer(buffer)
        return buffer.to_bytes()



from __future__ import annotations
from airtouch2.common.interfaces import Serializable


class Buffer(Serializable):
    """
    Layer of asbtraction on top of bytearray.
    Has a fixed size and enforces it's filled before it's serialized.
    Prohibits reading more than its size and before it's filled.
    """
    _data: bytearray
    _head: int = 0
    _tail: int = 0
    _mutable: bool = True

    def __init__(self, size: int):
        self._data = bytearray(size)

    def __len__(self):
        return len(self._data)

    def append_bytes(self, data: bytes) -> bool:
        """
        Return true and finalise if buffer was filled, raises BufferError if there
        is not sufficient room in the buffer or the buffer is already finalised.
        """
        if not self._mutable:
            raise BufferError("Buffer has been filled and is immutable")
        if (len(data) > len(self._data) - self._head):
            raise BufferError(
                "Buffer does not have enough room to append this data")
        for byte in data:
            self._data[self._head] = byte
            self._head += 1
        if self._head == len(self._data):
            self._mutable = False
            return True
        return False

    def append(self, object: Serializable) -> bool:
        """
        Return true and finalise if buffer was filled, raises BufferError if there
        is not sufficient room in the buffer or the buffer is already finalised.
        """
        return self.append_bytes(object.to_bytes())

    def to_bytes(self) -> bytes:
        if (self._mutable):
            raise BufferError(
                f"Buffer is not filled - {self._head}/{len(self._data)} bytes filled")
        return self._data

    def read_bytes(self, size: int) -> bytes:
        if (self._tail >= len(self._data)):
            raise BufferError("All data from this buffer has been read")
        if (self._tail >= self._head):
            raise BufferError("There is no remaining data to read")
        if (self._mutable):
            raise BufferError("Cannot read from incomplete buffer")
        start = self._tail
        self._tail += size
        return self._data[start:self._tail]

    def read_remaining(self) -> bytes:
        return self.read_bytes(self._head - self._tail)

    @staticmethod
    def from_bytes(data: bytes) -> Buffer:
        buffer = Buffer(len(data))
        buffer.append_bytes(data)
        return buffer




from airtouch2.common.Buffer import Buffer


def checksum(data: bytearray) -> int:
    sum: int = 0
    for b in data:
        sum += b
    sum = (sum % 256)
    return sum


def add_checksum_message_buffer(buffer: Buffer) -> None:
    buffer.append_bytes(checksum(buffer._data[:-1]).to_bytes(1, 'little'))



from __future__ import annotations
from typing import TYPE_CHECKING
from airtouch2.protocol.at2plus.messages.AcControl import AcControlMessage, AcSettings
from airtouch2.common.interfaces import Callback
if TYPE_CHECKING:
    from airtouch2.at2plus.At2PlusClient import At2PlusClient
from asyncio import Event
from airtouch2.protocol.at2plus.enums import AcFanSpeed, AcPower, AcSetMode, AcSetPower
from airtouch2.protocol.at2plus.messages.AcAbilityMessage import AcAbility
from airtouch2.protocol.at2plus.messages.AcStatus import AcStatus


class At2PlusAircon:
    """
    A class that represents a single airtouch2+ AC unit.

    An At2PlusAircon is not 'ready' until it's AcAbility has been retrieved.

    While unready, mode and fan speed setter calls cannot be made as the unit's supported modes are unknown.
    """

    def __init__(self, status: AcStatus, client: At2PlusClient):
        self.status: AcStatus = status
        self.ability: AcAbility | None = None
        self._ready: Event = Event()
        self._client: At2PlusClient = client
        self._callbacks: list[Callback] = []

    async def _set_power(self, power: AcSetPower):
        settings = AcSettings(self.status.id, power, AcSetMode.UNCHANGED, AcFanSpeed.UNCHANGED, None)
        await self._client.send(AcControlMessage([settings]))

    async def toggle(self):
        await self._set_power(AcSetPower.TOGGLE)

    async def turn_on(self):
        await self._set_power(AcSetPower.ON)

    async def turn_off(self):
        await self._set_power(AcSetPower.OFF)

    def is_on(self) -> bool:
        return self.status.power == AcPower.ON

    async def set_mode(self, mode: AcSetMode):
        settings = AcSettings(self.status.id, AcSetPower.UNCHANGED, mode, AcFanSpeed.UNCHANGED, None)
        await self._client.send(AcControlMessage([settings]))

    async def set_fan_speed(self, speed: AcFanSpeed):
        settings = AcSettings(self.status.id, AcSetPower.UNCHANGED, AcSetMode.UNCHANGED, speed, None)
        await self._client.send(AcControlMessage([settings]))

    async def set_setpoint(self, setpoint: float):
        settings = AcSettings(self.status.id, AcSetPower.UNCHANGED, AcSetMode.UNCHANGED, AcFanSpeed.UNCHANGED, setpoint)
        await self._client.send(AcControlMessage([settings]))

    async def wait_until_ready(self) -> None:
        await self._ready.wait()

    def add_callback(self, callback: Callback) -> Callback:
        self._callbacks.append(callback)

        def remove_callback() -> None:
            if callback in self._callbacks:
                self._callbacks.remove(callback)

        return remove_callback

    def _update_status(self, status: AcStatus):
        self.status = status
        for callback in self._callbacks:
            callback()

    def _set_ability(self, ability: AcAbility):
        self.ability = ability
        self._ready.set()



from __future__ import annotations
import asyncio
import logging
from typing import TYPE_CHECKING, Callable
from airtouch2.common.interfaces import Publisher, Callback, add_callback
if TYPE_CHECKING:
    from airtouch2.at2.At2Client import At2Client
from airtouch2.protocol.at2.enums import ACFanSpeed, ACBrand, ACMode
from airtouch2.protocol.at2.messages import ChangeSetTemperature, SetFanSpeed, SetMode, ToggleAc
from airtouch2.protocol.at2.messages.SystemInfo import AcInfo

_LOGGER = logging.getLogger(__name__)


class At2Aircon(Publisher):
    info: AcInfo

    def __init__(self, client: At2Client, info: AcInfo):
        self._client: At2Client = client
        self._callbacks: list[Callable] = []
        self.info = info

    def update(self, info: AcInfo) -> None:
        self.info = info

        for callback in self._callbacks:
            callback()

    def add_callback(self, callback: Callback) -> Callback:
        return add_callback(callback, self._callbacks)

    async def inc_dec_set_temp(self, inc: bool):
        await self._client.send(ChangeSetTemperature(self.info.number, inc))

    async def set_set_temp(self, new_temp: int):
        temp_diff = new_temp - self.info.set_temp
        inc = temp_diff > 0
        for i in range(abs(temp_diff)):
            await self.inc_dec_set_temp(inc)
            await asyncio.sleep(0.1)  # server doesn't like being spammed, 0.1s is faster than waiting for response.

    async def turn_off(self):
        await self._turn_on_off(False)

    async def turn_on(self):
        await self._turn_on_off(True)

    async def set_fan_speed(self, fan_speed: ACFanSpeed):
        if fan_speed in self.info.supported_fan_speeds:
            await self._client.send(SetFanSpeed(self.info.number, self.info.supported_fan_speeds, fan_speed))
        else:
            _LOGGER.warning(f"Cannot set fan speed to unsupported value {fan_speed}")

    async def set_mode(self, mode: ACMode):
        await self._client.send(SetMode(self.info.number, mode))

    async def _turn_on_off(self, on: bool):
        if self.info.active != on:
            await self._client.send(ToggleAc(self.info.number))

    def __str__(self):
        return str(self.info)



from airtouch2.protocol.at2plus.constants import Limits


def value_from_setpoint(setpoint: float | None) -> int:
    if setpoint:
        if not Limits.SETPOINT_MIN <= setpoint <= Limits.SETPOINT_MAX:
            raise ValueError(
                f'Setpoint must be between {Limits.SETPOINT_MIN} and {Limits.SETPOINT_MAX}')
        return int(setpoint * 10 - 100)
    return int(Limits.SETPOINT_MAX * 10 - 100 + 1)


def setpoint_from_value(value: int) -> float | None:
    setpoint = (value+100)/10
    if Limits.SETPOINT_MIN <= setpoint <= Limits.SETPOINT_MAX:
        return setpoint
    return None


def value_from_temperature(temp: float | None) -> int:
    if temp:
        if not Limits.TEMP_MIN <= temp <= Limits.SETPOINT_MAX:
            raise ValueError(
                f'Temperature must be between {Limits.TEMP_MIN} abd {Limits.TEMP_MAX}')
        else:
            return int(temp * 10 + 500)
    return int(Limits.TEMP_MAX * 10 + 501)


def temperature_from_value(val: int) -> float | None:
    temp: float = (val - 500)/10
    if Limits.TEMP_MIN <= temp <= Limits.TEMP_MAX:
        return temp
    return None
